/*
 * Copyright (c) 2015, Freescale Semiconductor, Inc.
 * Copyright 2016-2017 NXP
 * All rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

/* FreeRTOS kernel includes. */
#include <FreeRTOS.h>
#include <task.h>
#include <queue.h>
#include <timers.h>

/* Freescale includes. */
#include <fsl_device_registers.h>
#include <fsl_debug_console.h>
#include <fsl_lpspi.h>
#include <fsl_dcdc.h>

#include <pin_mux.h>
#include <clock_config.h>
#include <board.h>
#include <lpm.h>


#if configUSE_TICKLESS_IDLE == 2
#include "fsl_gpt.h"
#include <tickless_gpt.h>
#endif

/*******************************************************************************
 * Local variables
 ******************************************************************************/
static StaticTask_t m_idle_task_TCB;
static StackType_t m_idle_task_stack[configMINIMAL_STACK_SIZE];
static StaticTask_t m_timer_task_TCB;
static StackType_t m_timer_task_stack[configTIMER_TASK_STACK_DEPTH];

/*******************************************************************************
 * Code
 ******************************************************************************/

void vApplicationGetIdleTaskMemory(StaticTask_t **ppxIdleTaskTCBBuffer,
                                   StackType_t **ppxIdleTaskStackBuffer,
                                   uint32_t *pulIdleTaskStackSize) {
  /* Pass out a pointer to the StaticTask_t structure in which the
   * Idle task's state will be stored.
   */
  *ppxIdleTaskTCBBuffer = &m_idle_task_TCB;

  /* Pass out the array that will be used as the Idle task's stack. */
  *ppxIdleTaskStackBuffer = m_idle_task_stack;

  /* Pass out the size of the array pointed to by *ppxIdleTaskStackBuffer.
  Note that, as the array is necessarily of type StackType_t,
  configMINIMAL_STACK_SIZE is specified in words, not bytes. */
  *pulIdleTaskStackSize = configMINIMAL_STACK_SIZE;
}

void vApplicationGetTimerTaskMemory(StaticTask_t **ppxTimerTaskTCBBuffer,
                                    StackType_t **ppxTimerTaskStackBuffer,
                                    uint32_t *pulTimerTaskStackSize) {
  /* Pass out a pointer to the StaticTask_t structure in which the Timer
   * task's state will be stored.
   */
  *ppxTimerTaskTCBBuffer = &m_timer_task_TCB;

  /* Pass out the array that will be used as the Timer task's stack. */
  *ppxTimerTaskStackBuffer = m_timer_task_stack;

  /* Pass out the size of the array pointed to by *ppxTimerTaskStackBuffer.
   * Note that, as the array is necessarily of type StackType_t,
   * configTIMER_TASK_STACK_DEPTH is specified in words, not bytes.
   */
  *pulTimerTaskStackSize = configTIMER_TASK_STACK_DEPTH;
}


void vPreSleepProcessing( unsigned long ulExpectedIdleTime ) {
  GPC_DisableAllWakeupSource(GPC_CPU_MODE_CTRL);

  for (int i = 0; i < CLOCK_LPCG_NUM; i++) {
    CLOCK_LPCG_ControlByCpuLowPowerMode((clock_name_t)i, CM7_DOMAIN, 2, 0);
  }

  for (int i = 0; i < CLOCK_OSCPLL_NUM; i++) {
    CLOCK_OSCPLL_ControlByCpuLowPowerMode((clock_name_t)i, CM7_DOMAIN, 2, 0);
  }

  IOMUXC_GPR->GPR72 |= IOMUXC_GPR_GPR72_LPSPI6_IPG_DOZE_MASK;
  IOMUXC_GPR->GPR37 |= IOMUXC_LPSR_GPR_GPR37_LPSPI6_IPG_DOZE_MASK;
  EnableIRQ(SPI_DMA_IRQ);
  GPC_EnableWakeupSource(SPI_DMA_IRQ);

  IOMUXC_GPR->GPR71 |= IOMUXC_GPR_GPR71_LPI2C2_IPG_DOZE_MASK;
  IOMUXC_GPR->GPR36 |= IOMUXC_LPSR_GPR_GPR36_LPI2C2_IPG_DOZE_MASK;
  EnableIRQ(LPI2C_IMU_DMA_IRQ);
  GPC_EnableWakeupSource(LPI2C_IMU_DMA_IRQ);    

  // GPT (General Purpose Timer) interrupt
  EnableIRQ(GPT1_IRQn);
  GPC_EnableWakeupSource(GPT1_IRQn);

  // For debug
  LED_OFF(BLUE_LED_PIN);
}

void vPostSleepProcessing( unsigned long ulExpectedIdleTime ) {
  // For debug
  LED_ON(BLUE_LED_PIN);
}

/**
 * @brief Set break. Called by malloc.
 * @param incr[in] Increment end of data segment
 * @return Pointer to new end of memory
 */
void *_sbrk(int incr) {
  PRINTF("malloc() not supported\n");
  asm("bkpt");
}

void GPT1_IRQHandler(void) {
  vPortGptIsr();
}

GPT_Type *vPortGetGptBase(void) {
  return GPT1;
}

IRQn_Type vPortGetGptIrqn(void) {
  return GPT1_IRQn;
}

/*******************************************************************************
 * Init Task Definitions
 ******************************************************************************/
/* Task priorities. */
#define init_task_PRIORITY (configMAX_PRIORITIES - 2)
#define kInitStackSize     512   // 512 * 4 = 2048 bytes
static StackType_t m_init_task_stack[kInitStackSize];
static StaticTask_t m_init_task_TCB;

/*******************************************************************************
 * Code
 ******************************************************************************/
int main(void) {
  /* Init board hardware. */
  BOARD_ConfigMPU();
  BOARD_InitPins();

  clock_root_config_t rootCfg = {0};
  /* Configure LPUART1 using SYS_PLL2_CLK */
  rootCfg.mux = kCLOCK_LPUART1_ClockRoot_MuxSysPll2Out;
  rootCfg.div = 22;
  CLOCK_SetRootClock(kCLOCK_Root_Lpuart1, &rootCfg);

  /* Since core1 is not used and is always in reset status, force core1 requesting standby mode */
  GPC_STBY_CTRL->STBY_MISC |= GPC_STBY_CTRL_STBY_MISC_FORCE_CPU1_STBY_MASK;

  BOARD_BootClockRUN();
  BOARD_I2C_ConfigurePins();
  BOARD_InitDebugConsole();
  BOARD_Accel_I2C_Init();
  
  test_start();
  vTaskStartScheduler();
  for (;;)
      ;
}