/* FreeRTOS kernel includes. */
#include <FreeRTOS.h>
#include <task.h>
#include <timers.h>
#include <event_groups.h>

#include "fsl_debug_console.h"
#include "pin_mux.h"
#include "clock_config.h"
#include "board.h"
#include "IfSpi.h"
#include "imu.h"
#include "lpm.h"

#if configUSE_TICKLESS_IDLE == 2
#include "fsl_gpt.h"
#include <tickless_gpt.h>
#endif

/* Task priorities. */
#define test_task_PRIORITY       (configMAX_PRIORITIES - 4)
#define kTestStackSize           512   // 512 * 4 = 2048 bytes

#define SPI_DMA_RX_CHANNEL          0U
#define SPI_DMA_TX_CHANNEL          1U
#define SPI_PCS                     kLPSPI_Pcs0
#define SPI_PCS_XFER                kLPSPI_MasterPcs0
#define SPI_DMA_RX_REQUEST_SRC      kDmaRequestMuxLPSPI6Rx
#define SPI_DMA_TX_REQUEST_SRC      kDmaRequestMuxLPSPI6Tx
#define SPI_XFER_DONE_BIT (1 << 0)

#define LPI2C_IMU_DMA_IRQ DMA3_DMA19_IRQn
#define SPI_DMA_IRQ       DMA0_DMA16_IRQn

uint8_t m_write_buf[16*1024] = { 0x00 };
uint8_t m_read_buf[16*1024] = { 0x00 };
uint8_t m_read_buf_i2c[2*1024] = { 0x00 };

static edma_handle_t lpspiRxRegToRxDataHandle;
static edma_handle_t lpspiTxDataToTxRegHandle;
static StackType_t m_test_task_stack[kTestStackSize];
static StaticTask_t m_test_task_TCB;

bool spi_peripheral_init(void) {

  LPSPI_Enable(LPSPI6, false);

  // DMA MUX init.
  DMAMUX_Init(DMAMUX0);
  DMAMUX_SetSource(DMAMUX0, SPI_DMA_RX_CHANNEL, SPI_DMA_RX_REQUEST_SRC);
  DMAMUX_SetSource(DMAMUX0, SPI_DMA_TX_CHANNEL, SPI_DMA_TX_REQUEST_SRC);
  DMAMUX_EnableChannel(DMAMUX0, SPI_DMA_RX_CHANNEL);
  DMAMUX_EnableChannel(DMAMUX0, SPI_DMA_TX_CHANNEL);

  // EDMA init.
  edma_config_t userConfig;
  EDMA_GetDefaultConfig(&userConfig);
  userConfig.enableHaltOnError = false;
  EDMA_Init(DMA0, &userConfig);

  memset(&(lpspiRxRegToRxDataHandle), 0, sizeof(lpspiRxRegToRxDataHandle));
  memset(&(lpspiTxDataToTxRegHandle), 0, sizeof(lpspiTxDataToTxRegHandle));
  EDMA_CreateHandle(&lpspiRxRegToRxDataHandle, DMA0, SPI_DMA_RX_CHANNEL);
  EDMA_CreateHandle(&(lpspiTxDataToTxRegHandle), DMA0, SPI_DMA_TX_CHANNEL);

  // SPI config.
  lpspi_master_config_t spiLeadConfig;
  LPSPI_MasterGetDefaultConfig(&spiLeadConfig);
  spiLeadConfig.baudRate = SPI_TRANSFER_BAUDRATE;
  spiLeadConfig.bitsPerFrame = 8;
  spiLeadConfig.cpha = kLPSPI_ClockPhaseSecondEdge;
  spiLeadConfig.cpol = kLPSPI_ClockPolarityActiveHigh;
  spiLeadConfig.pcsToSckDelayInNanoSec = 160;
  spiLeadConfig.lastSckToPcsDelayInNanoSec = 100;
  spiLeadConfig.betweenTransferDelayInNanoSec = 100;
  spiLeadConfig.whichPcs = SPI_PCS;
  spiLeadConfig.pcsActiveHighOrLow = kLPSPI_PcsActiveLow;
  LPSPI_MasterInit(LPSPI6, &spiLeadConfig, SPI_CLK_FREQ);
  LPSPI_MasterTransferCreateHandleEDMA(LPSPI6, &g_m_edma_handle,
                                       spi_xfer_callback, NULL,
                                       &lpspiRxRegToRxDataHandle,
                                       &lpspiTxDataToTxRegHandle);
  NVIC_SetPriority(SPI_DMA_IRQ, 2);
  NVIC_EnableIRQ(SPI_DMA_IRQ);

  gEventGroup = xEventGroupCreateStatic(&xStaticEventGroup);
  if (gEventGroup == NULL) {
    PRINTF("Failed to create an event group for SPI\r\n");
    return false;
  }
  return true;
}

bool spi_read(const uint8_t *writeBuf, uint32_t writeBufSize, uint8_t *readBuf,
              uint32_t readBufSize) {
  bool ret = true;
  status_t status;

  do {
    lpspi_transfer_t spiXfer = {0};
    spiXfer.txData   = (uint8_t*) writeBuf;
    spiXfer.rxData   = readBuf;
    spiXfer.dataSize = readBufSize;
    spiXfer.configFlags =
        SPI_PCS_XFER | kLPSPI_MasterByteSwap | kLPSPI_MasterPcsContinuous;

    xEventGroupClearBits(gEventGroup, SPI_XFER_DONE_BIT);

    m_spi_transfer_completed = false;

    LPSPI_MasterTransferEDMA(LPSPI6, &g_m_edma_handle, &spiXfer);

    // Make this call blocking and wait for EDMA completion callback.
    EventBits_t uBits = xEventGroupWaitBits(gEventGroup, SPI_XFER_DONE_BIT,
                                            pdTRUE, pdFALSE,
                                            pdMS_TO_TICKS(IRQ_WAIT_TIMEOUT_MS));
    if ((uBits & SPI_XFER_DONE_BIT) != SPI_XFER_DONE_BIT) {
      ret = false;
    }
  } while(0);

  return ret;
}

static void test_task(void *pvUnused) {

#if configUSE_TICKLESS_IDLE == 2
  gpt_config_t gptConfig;
  GPT_GetDefaultConfig(&gptConfig);
  gptConfig.enableRunInDoze = true;
  /* Initialize GPT module */
  GPT_Init(GPT1, &gptConfig);

  /* Divide GPT clock source frequency by 1 inside GPT module */
  GPT_SetClockDivider(GPT1, 1);
  /* Enable GPT Output Compare1 interrupt */
  GPT_EnableInterrupts(GPT1, kGPT_OutputCompare1InterruptEnable);
  // Make sure specifying the IRQ priority other than 0.
  NVIC_SetPriority(GPT1_IRQn, 1);
  /* Enable at the Interrupt */
  EnableIRQ(GPT1_IRQn);  
#endif

  spi_peripheral_init();
  i2c_peripheral_init();

  int i = 0;
  while (1) {
    if (!spi_read(m_write_buf, 0, m_read_buf, sizeof(m_read_buf_spi))) {
      PRINTF("SPI XFER failed\r\n");
    }
    if (++i == 5) {
      i = 0;
      if (!i2c_read(m_read_buf_i2c, sizeof(m_read_buf_i2c))) {
        PRINTF("I2C XFER failed\r\n");
      }
    }

    vTaskDelay(pdMS_TO_TICKS(100));
  }
}

BaseType_t test_start(void) {
  TaskHandle_t xHandle;

  if ((xHandle = xTaskCreateStatic(test_task,
                                   "Test",
                                   kTestStackSize, NULL,
                                   test_task_PRIORITY,
                                   m_test_task_stack,
                                   &m_test_task_TCB)) == NULL) {
    PRINTF("[%s] %s() failed: 0x%08x\n", "TEST", "xTaskCreateStatic", -1);
    return pdFAIL;
  }
}

void spi_xfer_callback(LPSPI_Type *base, lpspi_master_edma_handle_t *handle,
                     status_t status, void *userData) {
  if (status == kStatus_Success) {
    m_spi_transfer_completed = true;
  }

  BaseType_t xHigherPriorityTaskWoken = pdFALSE;
  xEventGroupSetBitsFromISR(gEventGroup,
                            SPI_XFER_DONE_BIT,
                            &xHigherPriorityTaskWoken);
  portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}